#include <iostream>
#include <fstream>
#include <string>
#include <vector>
#include <cstdio>

#include "btypes.h"
#include "bfile.h"
#include "bdata.h"
#include "bparse.h"

BLFile *blfile;

std::ofstream outfile;

struct Variable {
  int typeID;
  int length;
  std::string type;
  std::string name;
  std::vector<float> floatValue;
  char charValue;
  std::vector<short> shortValue;
  std::vector<int> intValue;
  unsigned long pointerValue;
  std::string stringValue;
};

struct Structure {
  unsigned long ID;
  std::string name;
  bool alreadyExported;
  std::vector<Variable> variables;
};

std::vector<Structure> structures;

int FindStructure(int id)
{
  int i;

  for(i=0;i<structures.size();i++) {
    if(structures[i].ID==id)
      return i;
  }

  std::cerr << "fatal: unable to find structure by ID: " << id << std::endl;
  return 0;
  exit(-1);
}

Variable *FindVariable(int id, std::string name)
{
  int i;

  for(i=0;i<structures[id].variables.size();i++) {
    if(structures[id].variables[i].name==name)
      return &(structures[id].variables[i]);
  }

  std::cerr << "fatal: unable to find structure member by name:" << name << std::endl;
  std::cerr << "members are:" << std::endl;
  for(i=0;i<structures[id].variables.size();i++)
    std::cerr << structures[id].variables[i].name << std::endl;
  exit(-1);
}

void ConvertStruct(BLStruct *blstruct);

void ConvertVariable(int structNum, BLData *data)
{
  structures[structNum].variables.resize(structures[structNum].variables.size()+1);

  Variable *var=&(structures[structNum].variables[structures[structNum].variables.size()-1]);
  var->type=data->type_name;
  var->name=data->var_name;
  var->typeID=data->type;
  var->length=data->length;
  switch(data->type) {
  case BFLOAT:
    var->floatValue.resize(var->length);
    for(int i=0;i<var->length;i++)
      var->floatValue[i]=data->data.flt[i];
    break;
  case BINT:
    var->intValue.resize(var->length);
    for(int i=0;i<var->length;i++)
      var->intValue[i]=data->data.intg[i];
    break;
  case BSHORT:
    var->shortValue.resize(var->length);
    for(int i=0;i<var->length;i++)
      var->shortValue[i]=data->data.shrt[i];
    break;
  case BCHR:
    if(var->length==1)
      var->charValue=*data->data.chr;
    else 
      var->stringValue=std::string(data->data.chr).substr(0,var->length);
    break;
  case BPTR:
    var->pointerValue=*data->data.ptr;
    break;
  case BSTRC:
    var->pointerValue=(unsigned long)data->data.strc;
    ConvertStruct(data->data.strc);
    break;
  };
}

void ConvertStruct(BLStruct *blstruct)
{
  int last=structures.size();
  int i;

  structures.resize(structures.size()+1);
  structures[last].name=blstruct->struct_name;
  structures[last].ID=(unsigned long)blstruct;
  structures[last].variables.clear();
  structures[last].alreadyExported=false;

  for(i=0;i<blstruct->length;i++) {
    ConvertVariable(last,blstruct->data[i]);
  }
}

void ConvertStructs()
{
  BLStruct *temp;

  structures.clear();

  temp=blfile->struct_head;
  while(temp){
    ConvertStruct(temp);
    temp=temp->next;
  }
}

void OutputStruct(int numStruct);

void OutputVariable(Structure *structure, int i)
{
  int j;

  switch(structure->variables[i].typeID) {
  case BSTRC:
    OutputStruct(FindStructure(structure->variables[i].pointerValue));
    break;
  case BFLOAT:
    outfile << "  <Variable Name=\"" << structure->variables[i].name << "\" Type=\"float\" ";
    outfile << "Length=\"" << structure->variables[i].length << "\">" << std::endl;
    for(j=0;j<structure->variables[i].length;j++) {
      outfile << "    <Data Value=\"" << structure->variables[i].floatValue[j] << "\"/>" << std::endl;
    }
    outfile << "  </Variable>" << std::endl;
    break;
  case BSHORT:
    outfile << "  <Variable Name=\"" << structure->variables[i].name << "\" Type=\"short\" ";
    outfile << "Length=\"" << structure->variables[i].length << "\">" << std::endl;
    for(j=0;j<structure->variables[i].length;j++) {
      outfile << "    <Data Value=\"" << structure->variables[i].shortValue[j] << "\"/>" << std::endl;
    }
    outfile << "  </Variable>" << std::endl;
    break;
  case BINT:
    outfile << "  <Variable Name=\"" << structure->variables[i].name << "\" Type=\"integer\" ";
    outfile << "Length=\"" << structure->variables[i].length << "\">" << std::endl;
    for(j=0;j<structure->variables[i].length;j++) {
      outfile << "    <Data Value=\"" << structure->variables[i].intValue[j] << "\"/>" << std::endl;
    }
    outfile << "  </Variable>" << std::endl;
    break;
  case BCHR:
    if(structure->variables[i].length==1) {
      outfile << "  <Variable Name=\"" << structure->variables[i].name << "\" Type=\"char\">" << std::endl;
      outfile << "    <Data Value=\"" << (int)structure->variables[i].charValue << "\"/>" << std::endl;
      outfile << "  </Variable>" << std::endl;
    } else {
      outfile << "  <Variable Name=\"" << structure->variables[i].name << "\" Type=\"string\">" << std::endl;
      outfile << "    <Data Value=\"" << structure->variables[i].stringValue << "\"/>" << std::endl;
      outfile << "  </Variable>" << std::endl;
    }
    break;
  case BPTR:
    outfile << "  <Variable Name=\"" << structure->variables[i].name << "\" Type=\"ID\">" << std::endl;
    outfile << "    <Data Value=\"" << structure->variables[i].pointerValue << "\"/>" << std::endl;
    outfile << "  </Variable>" << std::endl;
    break;      
  }
}

void OutputStruct(int numStruct)
{
  int i, j;
  Structure *structure=&(structures[numStruct]);

  outfile << "<" << structure->name << " ID=\"" << structure->ID << "\">" << std::endl;

  for(i=0;i<structure->variables.size();i++) {
    OutputVariable(structure,i);
  }

  outfile << "</" << structure->name << ">" << std::endl;

  structure->alreadyExported=true;
}

void ExtractRemainingData()
{
  int i, j;

  for(i=0;i<structures.size();i++) {
    if(structures[i].alreadyExported) {
      continue;
    }
    OutputStruct(i);
  }
}

void ExtractMeshes()
{
  int i, j, k;
  Variable *temp;
  int meshID;
  int firstVertex;
  int numVertices;
  int firstFace;
  int numFaces;
  int firstUVFace;

  for(i=0;i<structures.size();i++) {
    if(structures[i].name=="Mesh") {
      meshID=i;

      outfile << "<Mesh ID=\"" << structures[meshID].ID << "\">" << std::endl;

      for(j=0;j<structures[meshID].variables.size();j++) {
	OutputVariable(&(structures[meshID]),j);
      }

      // Extract the number of vertices and faces:
      temp=FindVariable(meshID,"totvert");
      numVertices=temp->intValue[0];
      temp=FindVariable(meshID,"totface");
      numFaces=temp->intValue[0];

      // Extract the IDs of the first vertex and face:
      // XXX: The following code assumes that MVert, TFace and MFace data structures are alligned as arrays!
      temp=FindVariable(meshID,"*mvert");
      firstVertex=FindStructure(temp->pointerValue);
      temp=FindVariable(meshID,"*mface");
      firstFace=FindStructure(temp->pointerValue);
      temp=FindVariable(meshID,"*tface");
      firstUVFace=FindStructure(temp->pointerValue);

      // Extract the vertex coodinates:
      for(j=0;j<numVertices;j++) {
	OutputStruct(firstVertex+j);
      }

      // Extract faces and UV coordinates:
      for(j=0;j<numFaces;j++) {
	OutputStruct(firstFace+j);
	OutputStruct(firstUVFace+j);
      }

      outfile << "</Mesh>" << std::endl;

      structures[meshID].alreadyExported=true;
    }
  }
}

int main(int argc, char *argv[])
{
  if(argc!=3) {
    std::cerr << "usage: " << argv[0] << " [blender file] [output file]" << std::endl;
    return -1;
  }

  blfile=BLOpenFile(argv[1]);
  if(!blfile) {
    std::cerr << argv[0] << ": unable to open file \"" << argv[1] << "\"" << std::endl;
    return -1;
  }

  outfile.open(argv[2]);
  if(!outfile) {
    std::cerr << argv[0] << ": unable to open file \"" << argv[2] << "\"" << std::endl;
    return -1;
  }

  BLParseFile(blfile);
  ConvertStructs();

  std::cout << "read " << structures.size() << " structures from file" << std::endl;

  outfile << "<BlendFile>" << std::endl;
  ExtractMeshes();
  ExtractRemainingData();
  outfile << "</BlendFile>" << std::endl;
}
